#********************************************************************************
# Administration of qemu virtual machines to use with qemu-init                 *
#                                                                               *
#  Author: Kai Peter, ©2016-                                                    *
# Version: 0.50                                                                 *
# License: ISC                                                                  *
#                                                                               *
# Description: Control of all administrative functions of 'qemu-init'.          *
#********************************************************************************
VERSION=`cat $0 | grep -m 1 'Version' | cut -d\   -f3`       # get version number
ME=`basename $0`
CFF="REGFILE"

PATH=QLIBDIR:`dirname $0`:$PATH
declare -i DBG=0

showHelp() {
  echo -e "Usage qemu-adm <action> [ <name> ] [ <option> ]"
  echo -e ""
  echo -e "Actions:"
  echo -e "   qemu-adm [ -a|-r|-u ] <name>" # [ -f ]"
  echo -e "   qemu-adm [ -S|-D|-R|-m|-s|--config ] <name>"
  echo -e "   qemu-adm -c <command> <name>"
  echo -e "   qemu-adm [ -l ] | [ -h ] | [ -V ]"
  echo -e ""
  echo -e "   -a | --autostart <name> \t change (revert) the AUTOSTART option"
  echo -e "   -c | --cmd <command> <name> \t run qemu monitor instruction <command> for <name>."
  echo -e "                           \t <command> have to be quoted properly if necessary."
  echo -e "   --config <name>         \t show configuration of virtual machines from rc file"
  echo -e "   -D | --stop <name>      \t stop virtual machine <name>"
  echo -e "   -h | --help             \t show this help screen"
  echo -e "   -l | --list             \t list all registered virtual machines"
  echo -e "   -m | --monitor <name>   \t connect to the qemu monitor via tcp"
  echo -e "   -r | --register <name>  \t register a virtual machine"
  echo -e "   -R | --restart <name>   \t restart virtual machine <name>"
  echo -e "   -S | --start <name>     \t start virtual machine <name>"
  echo -e "   -s | --status <name>    \t show runtime configuration of <name>"
  echo -e "   -u | --unregister <name>\t unregister virtual machine <name>"
  echo -e "   -V | --version          \t show version info and exit"
  echo -e ""
  echo -e "Options:"
  echo -e "   --cdrom [<file>]        \t eject cdrom drive and mount <file> to first cd drive"
  echo -e "                           \t w/o <file> the first cdrom will be ejected only"
  echo -e "   -f                      \t a file operation depending on action"
  echo -e ""
}

error() {
  printf "\033[1merror: \033[0m"
  case $1 in
    1) echo "Invalid directory!";;
    2) echo "Couldn't find template file!";;
    3) echo "Invalid number of arguments.";;
  # registration errors
   10) echo "An absolute path is required!";;
   11) echo "A 'minus' or 'dot' is not allowed.";;
   12) echo "A VM with name '$2' is registered already!";;
   13) echo "Couldn't create config file $2!";;
   16) echo "Couldn't un-register '$2'.";;
   18) echo "Not a registered VM: $2";;
   20) echo "$2privilege escalation required!";;
  # monitor command issues
   30) echo "No cdrom drive found.";;
  esac
  exit $1
}

warning() {
  printf "\033[1mwarning: \033[0m"
  case $1 in
   30) echo "Invalid file: $FLAG_CDROM";;
  esac
}

list() {
  printf "\033[93mRegistered VM's:\033[0m\n"
  [ ! -f "$CFF" ] && return
  local i
  printf " \033[4m\033[1m%-18s %-40s %-12s \033[0m\n" \
         "Name" "|Location" "|Autostart"
  for L in `cat "$CFF" | grep -v '^#' | grep 'qemu.rc'`
  do
    # The short name is part of the first field of 'REGFILE':
    # We split this field using "/" as delimiter and  count the number of "/"'s -
    # the (splitted) part after the latest "/" contains the VM's name (w/o path).
    i=`expr $(echo $L | tr -cd '/' | wc -c) + 1`
    printf " %-18s" " `echo "$L" | cut -d: -f1 | cut -d- -f1 | cut -d/ -f${i}`"
    printf " %-40s" "| `echo "$L" | cut -d: -f1`"
    printf " %-12s" "| `echo "$L" | cut -d: -f2`"
    printf " %-12s" "  `echo "$L" | cut -d: -f3`"      # not used yet
    printf " %-12s" "  `echo "$L" | cut -d: -f4`"      #      "
    printf "\n"
  done | sort
}

register() {
  local VMRC=$1
  # the full path is required to register
  echo $VMRC | grep '^/' >/dev/null || error 10
  # check for suffix '-qemu.rc' and remove it if exists
  echo $VMRC | grep '-qemu.rc$' >/dev/null && \
    VMRC=`echo $VMRC | sed 's/\-qemu.rc$//g'`
  # check for additional minus and/or dot in the (unique) short name
  echo `basename $VMRC` | grep '[-.]' >/dev/null && error 11
  # split into 't'arget 'folder' and 's'hort 'n'ame
  local TF=`dirname $VMRC` ; local SN=`basename $VMRC`
  # the registered short name have to be unique
  grep "$SN" "$CFF" >/dev/null && error 12 "$SN"
  # create target folder if $TF is valid
  [ "$TF" ] && mkdir -p "$TF"

  # TODO: if folder exists already --> check for write access ???

  # create <name>-qemu.{rc,inc} from template if not exists
  if [ ! -f "$TF/$SN-qemu.rc" ] ; then
    cat "QETCDIR/template-qemu.rc" | \
    sed "s/[<]name[>][\"]/$SN\"/g" | \
    sed "s/[<]name-sda[>]/$SN-sda/g" \
    > "$TF/$SN-qemu.rc" || error 13 "$TN/$SN-qemu.rc"

    # user's config file template
    [ -f "QETCDIR/template-user.cfg" ] && \
      cp -n "QETCDIR/template-user.cfg" "$TF/$SN-user.cfg" || \
      printf "\0331mWarning: Couldn't create $SN-user.cfg!\033[0m\n"
  fi

  [ "$DBG" -gt 0 ] && echo Target folder: $TF && echo Unique name: $SN

  # create hd image
  if [ "$FLAG_F" ] && [ ! -f "$TF/$SN-qemu.img" ] ; then
    local QEMU_IMG
    QEMU_IMG=`which qemu-img 2>/dev/null`
    [ "$QEMU_IMG" ] && echo "creating disk image ..." && \
      ${QEMU_IMG} create -f qcow2 $TF/$SN-qemu.img 30G || \
      printf "\033[1mWarning: Couldn't create hd image!\033[0m\n"
  fi

  [ "$DBG" -gt 0 ] && return
  # register VM
  echo "$VMRC-qemu.rc:OFF::" >> "$CFF"
  [ $? = 0 ] && echo "Successful registered $SN ($TF/$SN-qemu.rc)"
}

unregister() {
  local VMRC=$1
  # check for suffix '-qemu.rc' and remove it if exists
  echo $VMRC | grep '-qemu.rc$' >/dev/null && \
    VMRC=`echo $VMRC | sed 's/\-qemu.rc$//g'`

  local SN=`basename $VMRC`    # now get the 's'hort 'n'ame $SN

  # the short name $SN starts after the last slash of its path and has the suffix
  # '-qemu.rc', thus the 'grep "/$SN-qemu.rc"' is unique.
  grep "/$SN-qemu.rc" "$CFF" >/dev/null && \
  cat "$CFF" | grep -v $VMRC > "$CFF.tmp"
  [ $? = 0 ] && sudo mv "/tmp/`basename $CFF`.tmp" "$CFF" && \
  sudo chmod 0664 "$CFF" && sudo chown QUSER.QGROUP "$CFF" && \
  echo "$VMRC successful unregistered." && exit 0

  error 16 "$SN"    # ... or print an error if something went wrong
}

autostart() {
  # 'root' is required to change this, otherwise users could/will lose control of
  # their VM's if started at boot time (they are owned by root than).
  [ ! `id -u` = 0 ] && error 20 "change autostart: "

  local VMRC=$1
  # check for suffix '-qemu.rc' and remove it if exists
  echo $VMRC | grep '-qemu.rc$' >/dev/null && \
    VMRC=`echo $VMRC | sed 's/\-qemu.rc$//g'`

  local SN=`basename $VMRC`    # now get the 's'hort 'n'ame $SN

  grep "/$SN-qemu.rc" "$CFF" >/dev/null || error 18 "$SN"

  while IFS= read L ; do
    echo "$L" | grep "$SN-qemu.rc" > /dev/null
    if [ $? = 0 ] ; then
      F1=`echo "$L" | cut -d: -f1`
      F2=`echo "$L" | cut -d: -f2`
      F2=`echo "$F2" | tr '[A-Z]' '[a-z]'`
      [ "$F2" = "on"  ] && F2=1
      [ "$F2" = "yes" ] && F2=1
      [ "$F2" = "y"   ] && F2=1
      [ "$F2" = "1"   ] && F2=1
      [ "$F2" = "1" ] && F2=OFF || F2=ON
      L="$F1:$F2:::"
    fi
    echo "$L" >> "/tmp/`basename $CFF`.tmp"
  done < "$CFF"
  mv "/tmp/`basename $CFF`.tmp" "$CFF" && \
  chmod 0664 "$CFF" && chown QUSER.QGROUP "$CFF" && \
  echo "Switched AUTOSTART of $SN to: $F2"
}
#--------------------------------------------------------------------------------
#***** qemu monitor commands ****************************************************
cdrom(){
  local NAME=$1
  local DEVICE
  # we use the first cd drive which we found
  DEVICE=`qemu-cmd -c "info block" $NAME | grep -m 1 '\-cd'`
  DEVICE=`echo $DEVICE | cut -d\  -f1 | sed 's/://g'`
  [ ! "$DEVICE" ] && error 30

  # eject cdrom, unset FLAG_CDROM to prevent 'invalid file warning' and return
  [ "$FLAG_EJECT" ] && qemu-cmd -c "eject $DEVICE" "$NAME" >/dev/null && \
      FLAG_CDROM="" && return

  qemu-cmd -c "change $DEVICE $FLAG_CDROM" "$NAME" >/dev/null
  [ "$DBG" -gt 0 ] && echo -n "CD: " && \
    qemu-cmd -c "info block" "$NAME" | grep "^$DEVICE" | awk '{ print $3; };'
}
reset() {
  local NAME=$1

  # if FLAG_CDROM starts with a minus sign then it doesn't contain a file but the
  # next parameter - without an argument <file> it will be ejected.
  echo "$FLAG_CDROM" | grep '^\-' >/dev/null && FLAG_CDROM="eject"
  [ "$FLAG_CDROM" = "eject" ] && FLAG_EJECT="eject" && cdrom "$NAME"

  # insert <file> at the first cdrom if <file> exists
  [ -f "$FLAG_CDROM" ] && cdrom "$NAME"

  qemu-cmd -q -c "system_reset" "$NAME" >/dev/null
  exit $?
}
#--------------------------------------------------------------------------------
#***** begin processing *********************************************************
setsudo() {
  unset QEMUINIT_SUDO
  if [ `id -u` -gt 0 ] ; then
	getent group Q_GROUP | grep `id -un` >/dev/null
	[ $? = 0 ] && export QEMUINIT_SUDO="sudo"
  fi
}

PARAMS=$@                  # a string with all command line options
OPTARG=($PARAMS)           # an array with all the arguments
index=0                    # loop index
unset ACTION
unset DoExec

for ARG in $PARAMS ; do
  index=`expr $index + 1`        # increment the index
  case "$ARG" in
    # internal actions
  -a|--autostart) ACTION="autostart ${OPTARG[index]}";;
       -h|--help) showHelp ; exit 0;;
       -l|--list) ACTION="list";;                      # list all registered VM's
   -r|--register) ACTION="register ${OPTARG[index]}";;
         --reset) ACTION="reset ${OPTARG[index]}";;    # perform a warm start
 -u|--unregister) ACTION="unregister ${OPTARG[index]}";;
    -V|--version) echo "$ME-$VERSION" ; exit 0;;
    # external actions
       -D|--stop) ACTION="qemu-init ${OPTARG[index]} stop";;           # shutdown
      -S|--start) ACTION="sudo QLIBDIR/qemu-init"
                  ACTION="$ACTION ${OPTARG[index]} start";;            # start
        --config) ACTION="qemu-init ${OPTARG[index]} config" ;;        # config
    -R|--restart) ACTION="qemu-init ${OPTARG[index]} restart" ;;       # restart
     -s|--status) ACTION="qemu-init ${OPTARG[index]} status"           # status
                  # using 'qemud status' to show status for all virtual machines
                  if [ ! ${OPTARG[index]} ] || [ ${OPTARG[index]} = "all" ]
                     then setsudo ; ACTION="$QEMUINIT_SUDO qemud status" ; fi ;;
    -m|--monitor) # Warning!!! NEVER put 'qemu-mon' in the background!  Otherwise
                  # the tcp socket will not be closed after exiting qemu monitor.
                  ACTION="qemu-mon ${OPTARG[index]}";;
        -c|--cmd) # 'qemu-cmd' is special: it needs exactly two arguments and the
                  # first one (here: "$2") _COULD_ be quoted.  The easiest way is
                  # to pipe through the parameters of qemu-adm. Thus we allow two
                  # arguments for 'qemu-adm -c' only.
                  [ $# -gt 3 ] && error 3               # check number of options
                  qemu-cmd -c "$2" "$3" | grep -v '^(qemu)' | grep -v '^QEMU '
                  exit $? ;;
    # options (flags)
              -f) FLAG_F="f";;
          --test) echo "key $ARG value ${OPTARG[index]}";;
         --cdrom) FLAG_CDROM="${OPTARG[index]}"
                  [ ! "$FLAG_CDROM" ] && FLAG_CDROM="eject";;
  esac
  [ ! "$DoExec" ] && [ "$ACTION" ] && DoExec="$ACTION"    # the first ACTION wins
done
# simple warning for arguments ignored by 'qemu-init' or 'qemu-mon'
echo "$DoExec" | grep 'qemu-' >/dev/null && [ $# -gt 2 ] && \
   echo "Warning: ignoring gratuitous arguments!"

[ "$DBG" -gt 2 ] && echo "finished"
[ "$DoExec" ] && ${DoExec} ; exit $?
