#********************************************************************************
# Start/stop qemu virtual machines                                              *
#                                                                               *
#  Author: Kai Peter, ©2016-2020                                                *
# Version: 0.84                                                                 *
# License: ISC                                                                  *
#********************************************************************************
VERSION=`cat $0 | grep -m 1 'Version' | cut -d\   -f3`       # get version number
ME=`basename $0`                                             # set program name
CFF="REGFILE"

# since cfg template version 0.58 the QEMUARCH variable should be used - or we do
# fallback to the old deprecated definition (which is static and basically wrong)
# !!! This is for backwards compat with older config files only -> deprecated !!!
[ "$TARGET_ARCH" ] && QEMUBIN="$TARGET_ARCH" || QEMUBIN="qemu-system-`uname -m`"

# we use this string for '--help' and as an error message
USAGE_MSG="Usage: "`basename $0`" <VM> start|stop|restart|status|config"

error() {
  # beware that the parameters here are independent from 'basename $0' one's
  ERRMSG1="Invalid call: use qemu-init!"
  ERRMSG2="$USAGE_MSG"
  ERRMSG3="Invalid name $2!"
  ERRMSG4="Not a registered VM: $2!"
  ERRMSG5="Can't change into working directory $WORKDIR!"
  ERRMSG6="Couldn't read config file!"
  ERRMSG7="Seems another instance of $FULLNAME is running already!"
  ERRMSG13="Couldn't start VM - check qemu options (see $2-qemu.log)!"
  ERRMSG14="Couldn't shutdown VM - couldn't connect to qemu monitor!"
  #
  printf "\033[1m  error: $(eval echo \"\$$(echo ERRMSG$1\"))\033[0m\n"
  exit $1
}
result() {
  C=`tput cols` ; C=$(($C-4))   # calculate cursor position (column)
  printf "\033[${C}G"           # move cursor to column ${C} in actual line
  [ "$EC"  = "B" ] && printf "\033[1m\033[93m%s\033[0m\n" "--" && return # BOOTDELAY
  [ "$EC"  = "0" ] && printf "\033[32m%s\033[0m\n" "OK"
  [ "$EC" != "0" ] && printf "\033[1m\033[91m%s\033[0m\n" "!!"
}

setports() {
  # get free port for monitor - have to be set before including config file
  [ -f /etc/qemu/monitor.conf ] && . /etc/qemu/monitor.conf || PORT=""
  [ "$PORT" = 0 ] && PORT=""          # setting PORT=0 disables monitor/vnc
  [ "$PORT" ] && MONPORT="$PORT"
  until [ ! "$MONPORT" ] || [ "$MONPORT" -gt 65000 ] ; do
    ps ax | grep "qemu-system" | grep ":$MONPORT," 2>&1 >/dev/null
    EC=$? ; [ "$EC" -ne 0 ] && break
    MONPORT=$(($MONPORT+1))
  done
  [ ! "$VNCPORT" ] && [ "$PORT" ] && VNCPORT=$(($MONPORT-$PORT))
}

#********************************************************************************
# build dynamic options for drives and NIC's as well control the vm's behavior
drives() {
  local i=0
  until [ ! "$DSK1" ] ; do
    i=$(($i+1))
    TMP=$(eval echo \"\$$(echo DSK$i)\")   # POSIX
    [ "$TMP" ] && DRIVES="$DRIVES -drive file=$TMP,snapshot=off" || break
  done
}
cdroms() {
  local i=0 ; CDROMS=""
  until [ ! "$CDR1" ] ; do
    i=$(($i+1))
    TMP=$(eval echo \"\$$(echo CDR$i)\")  # POSIX
    [ "$TMP" ] && CDROMS="$CDROMS -cdrom $TMP" || break
  done
  [ "$CDROMS" ] && DRIVES="$DRIVES $CDROMS"    # append to DRIVES if exists
}
netifs() {
  NETSCRIPTS="script=QLIBDIR/qemu-brup,downscript=QLIBDIR/qemu-brdown"
  local i=0
  until [ ! "$NIC1" ] ; do
    j=$i ; i=$(($i+1))
    MAC=`hexdump -n2 -e'"DE:AD:BE:EF" 2/1 ":%02X"' <<< $(sha256sum <<< "$NAME$i")`
    NETPARAMS="netdev=net$j,mac=$MAC -netdev tap,id=net$j,$NETSCRIPTS"
    TMP=$(eval echo \"\$$(echo NIC$i)\")     # 'eval' is POSIX
    [ "$TMP" ] && NET="$NET -device $TMP,$NETPARAMS" || break
  done
}
custopts() {
  # read customer options from file (sequential)
  local VAR_NAMES
  local VAR_VALUES
  local L t
  while read LINE ; do
    L=`echo $LINE | grep -v '^#' | grep '='`        # read valid lines
    if [ "$L" ] ; then                              # and if it's not empty
      t=`echo $L | cut -d= -f1`                     # save var name temporyrily
      if [ "$t" != "CUST_OPTS" ] ; then             # CUST_OPTS is not allowed
        # add variable name to space delimited string
        VAR_NAMES=`echo $VAR_NAMES $t | tr ' ' '\n' | uniq`
        # get tthe variable value (w/o comments)
        t=`echo $L | cut -d\= -f2- | cut -d\# -f1`  # add values to VAR_VALUES
        VAR_VALUES=`echo $VAR_VALUES $t | tr ' ' '\n'`
      fi
    fi
  done < "$CUSTOM"
  # now clean up the VAR_VALUES string
  VAR_VALUES=`echo $VAR_VALUES | sed 's/\"//g'`     # remove " from VAR_VALUES
  for t in `echo "$VAR_NAMES"` ; do
    # remove all '$' and VAR_NAMES from $VAR_VALUES
    VAR_VALUES=`echo $VAR_VALUES | sed 's/[$]//g' | sed "s/$t//g"`
  done

  USROPTS="$VAR_VALUES"
  return
}
options() {      # consolidate options for functions start() and config() -------
  # check for custom qemu parameters and consolidate them
  [ "$CUSTOM" ] || CUSTOM=`echo $VMRC | sed 's/qemu.rc$/user.cfg/g'`
  [ -f "$CUSTOM" ] && custopts

  # option "-daemonize" have to be disabled explizit by 'DAEMON=0'
  [ "$DAEMON" ] && [ "$DAEMON" = 0 ] && DAEMON="" || DAEMON="-daemonize"

  # check QEMU_DISPLAY variable
  VIDEO=""
  if [ "$QEMU_DISPLAY" ] ; then
    QEMU_DISPLAY=`echo $QEMU_DISPLAY | tr 'A-Z' 'a-z'`
    [ "$QEMU_DISPLAY" = "vnc"  ] && VIDEO="-display vnc=:$VNCPORT" \
      && VIDEO="$VIDEO $MOUSE"
      #                ^^^^^^ -> perhaps this _could_ be a duplicate qemu option!
    [ "$QEMU_DISPLAY" = "none" ] && VIDEO="-display none"
  fi

  # just put the (more or less static) options from config file together
  OPTS="$CPU $DAEMON $BOOT $LOCAL $VIDEO"

  ip -6 addr 2>&1 >/dev/null   # check for IPv6
  EC=$? ; [ "$EC" = 0 ] && IPV6=1 || IPV6=0
  [ $MONPORT ] && OPTS="$OPTS -monitor tcp:localhost:$MONPORT,server,nowait,ipv4"
  [ "$IPV6" = 1 ] && [ $MONPORT ] && \
      OPTS="$OPTS -monitor tcp:localhost:$MONPORT,server,nowait,ipv6"
  # control monitor
  CTLPORT=$(($MONPORT+100))
  [ $MONPORT ] && OPTS="$OPTS -monitor tcp:localhost:$CTLPORT,server,nowait,ipv4"
  [ "$IPV6" = 1 ] && [ $MONPORT ] && \
      OPTS="$OPTS -monitor tcp:localhost:$CTLPORT,server,nowait,ipv6"

  drives ; cdroms ; netifs      # build these ones ...
  OPTS="$OPTS $DRIVES $NET"     # ... and add dynamically options ;)
  [ "$USROPTS" ] && OPTS="$OPTS $USROPTS"
  # CUST_OPTS is deprecated! It will not work with <name>-user.cfg!
  [ "$CUST_OPTS" ] && OPTS="$OPTS $CUST_OPTS"
}

#********************************************************************************
check() {                        # return code: 0 if VM is running or 1 otherwise
  # if VM is running w/o pidfile - bummer!
  [ ! -f $PIDFILE ] && EC=1 && return "$EC"
  # remember that the 'FULLNAME' have to be unique always
  ps ax | grep `head -1 $PIDFILE` | grep "$FULLNAME" 2>&1 >/dev/null
  EC=$? ; [ "$EC" -ne 0 ] && EC=1    # usually '$?' is 1 on error, but who knows
  [ "$EC" -ne 0 ] && rm "$PIDFILE"   # remove pidfile (the PID is invalid anyway)
  return "$EC"
}

config() {                  # read and display the configuration from config file
  setports                  # results will be overwritten if VM is up and running
  # we assume that the pidfile exists always if the VM is running - see check()
  check
  if [ "$EC" -eq 0 ] ; then
    [ -f "$PIDFILE" ] && MONPORT=`grep '^MON' "$PIDFILE" | cut -d= -f2`
    [ -f "VNCFILE" ] && \
      VNCPORT=`grep -i "$NAME" VNCFILE | cut -d: -f3`
    # calculate vnc offset to display option '-display vnc=:X' (see qemu docu)
    [ "$VNCPORT" ] && [ "$PORT" ] && VNCPORT=$(($MONPORT-$PORT))
  fi
  options
  # options 'pidfile' and 'name' differs against function 'start'
  OPTS="$OPTS -pidfile $PIDFILE -name $FULLNAME"
  printf "Using \033[1m$VMRC\033[0m:\n"
  echo `which $QEMUBIN`
  # display one option per line - use ';' as field-separator
  TMP=`echo $OPTS | sed 's/ -/;/g' | sed 's/^-//g'`   # remove leading '-' too
  awk -F";" '{for (i=1;i<=NF;i++) printf("\t-%s\n", $i)}' <<< "$TMP"
  printf "\n\033[1m" ; status ; printf "\033[0m"
  # print next message if the VM isn't running
  [ "$EC" -ne 0 ] && [ ! -f "$PIDFILE" ] && \
    echo "(tcp ports are estimated as it would be right now!)"
  return 0
}

start() {
  check ;  [ "$EC" -eq 0 ] && echo -n "$FULLNAME: " && error 7  # already running

  [ -f "$PIDFILE" ] && rm "$PIDFILE"   # see issue 226
  echo -n "Starting $FULLNAME ..."
  #******************************************************************************
  # The BOOTDELAY option will be used with qemud only.  It have to be an integer,
  # the value have to be greater than 0. Invalid values will be converted to 0.
  [ "$QEMUINIT" ] && declare -i INT="$BOOTDELAY"        # check for integer value
  if [ "$INT" ] && [ "$INT" -gt 0 ] ; then
    # remove $QEMUINIT from environment and  start a new instance in background -
    # ... and yes, it works perfect! And it is very simple, isn't it? :)
    unset QEMUINIT ; ( sleep "$INT" && "$ME" "$VMRC" start >/dev/null & )
    EC="B" && echo -n " ($INTs delayed)" && result      #  if it fails - bummer!
    exit 0
  fi

  [ "$PRECMD" ] && `$PRECMD`     # if it fails - bummer
  setports
  options
  # options 'pidfile' and 'name' HAVE TO be added here explicit!
  "$QEMUBIN" -name "$FULLNAME" -pidfile $PIDFILE $OPTS 2>$LOGFILE
  EC=$? && result
  [ $EC -ne 0 ] && error 13 "$WORKDIR/`echo $NAME | tr 'A-Z' 'a-z'` (code $EC)"
  chown "$SUDO_USER".QGROUP $PIDFILE && chmod g+rw $PIDFILE
  [ "$MONPORT" ] && echo "MON=$((MONPORT))" >> "$PIDFILE"
  qemu-vnc
}

down() {
  TMP=$(echo `basename $VMRC` | cut -d\- -f1)   # use the 'short' name
  # do not allow shutdown for unregistered VM's
  grep "$TMP" "$CFF" > /dev/null
  [ "$?" -ne 0 ] && error 4 "$TMP"

  echo -n "Shutting down $FULLNAME ..."
  check
  [ "$EC" -ne 0 ] && result && echo "  $FULLNAME is not running!" && return

  CTLPORT=`grep '^MON' "$PIDFILE" | cut -d= -f2`
  CTLPORT=$(($CTLPORT+100))
  [ ! "$CTLPORT" ] && error 14 && return
  qemu-cmd -c "system_powerdown" "$TMP" 2>&1 >/dev/null
  [ "$STOPDELAY" ] && declare -i INT="$STOPDELAY"
  [ "$INT" ] && [ "$INT" -gt 0 ] && STOPDELAY="$INT" || STOPDELAY=40
  i=1   # check if VM is down (default: timeout 40 seconds)
  until [ $i -gt "$STOPDELAY" ] ; do
    qemu-cmd -q -c "info status" "$TMP" 2>&1>/dev/null
    EC=$?
    [ "$EC" -ne 0 ] && i=$(($i+$STOPDELAY+1))
    echo -n "." ; sleep 1
    i=$(($i+1))
  done
  # we have to reverse the return codes here
  [ "$EC" = 0 ] && EC=1 || EC=0
  result $EC
  [ "$EC" = 0 ] && rm -f $PIDFILE      # remove pidfile on successful shutdown
  qemu-vnc
}

restart() { down && sudo QLIBDIR/qemu-init "$VMRC" start; }

status() {
  # show runtime config only if 'qemu-init <name> status' is invoked directly
  if [ ! "$QEMUINIT" ] && [ ! ${ACTION} = "config" ] ; then
    VMRC=$(echo `basename $VMRC` | cut -d\- -f1)       # use the 'short' name
    TMP=`ps ax | grep $VMRC | grep qemu-system | sed 's/ -/;/g'`
    if [ "$TMP" ] ; then
      printf "$QEMUBIN\n"
      awk -F";" '{for (i=2;i<=NF;i++) printf("\t-%s\n", $i)}' <<< "$TMP"
      echo
    fi
  fi

  check
  if [ "$EC" -eq 0 ] ; then
     echo -e "$FULLNAME \033[32mis running\033[0m with PID `head -1 $PIDFILE`"
     return
  fi
  echo -e "qemu VM $FULLNAME is \033[93mnot running\033[0m"
  [ -f "$PIDFILE" ] && rm -f "$PIDFILE"
}

# as of now: don't allow direct call of links (.rc file) ************************
# Deprecated: For future version the '.rc-file' is no longer a link to qemu-init,
# indead it will be a regular config file copied from template-qemu.rc.
if [ "$ME" != "qemu-init" ] ; then
  [ ! "$QEMUINIT" ] && error 1  # this is set if 'qemu-init' is called by 'qemud'
  VMRC=$0
fi

case $1 in           # ----- the basic options every program should have --------
  -h|--help)    echo "$USAGE_MSG" ; exit 0;;
  -V|--version) echo "$ME-$VERSION" ; exit 0;;
esac

case $2 in
    start) ACTION="start";;
     stop) ACTION="down";;
  restart) ACTION="restart";;
   status) ACTION="status";;
   config) ACTION="config";;
        *) error 2;;
esac

#********************************************************************************
# There are 2 command line parameters:
#   $1 - name of the VM or rc file (symlink) with absolute path
#   $2 - action: start|stop|restart|status|config
echo $1 | grep ^/ 2>&1 >/dev/null         # check for .rc file with absolute path
if [ "$?" = 0 ] ; then                    # ++ (required for unregistered VM's)
  echo "$1" | grep '.rc$' >/dev/null || error 3 $1     # check for suffix '.rc'
  [ -f "$1" ] && VMRC="$1"                             # check if .rc file exists
  [ ! "$VMRC" ] && error 3 $1
 else
  # we allow to start - and only to start - an unregistered VM
  # 'grep $1-qemu.rc': the name have to be '<name>-qemu.rc'
  VMRC=$(grep $1-qemu.rc "$CFF" 2>/dev/null)
  VMRC=$(echo $VMRC | cut -d: -f1)
  [ ! "$VMRC" ] && error 4 $1
fi

WORKDIR=`dirname $VMRC` ; [ -d "$WORKDIR" ] && cd "$WORKDIR" || error 5

# Deprecated: using the '.cfg' file is for backwards compat only
CFGFILE=$(echo `basename $VMRC` | cut -d. -f1).cfg
[ -f "$CFGFILE" ] && \
  echo -e "\033[1mWarning! Usage of config file $CFGFILE is deprecated!\033[0m"
# Since 0.78: the '.rc' file is no longer a link to qemu-init but the base config
[ ! -f "$CFGFILE" ] && CFGFILE=$(echo `basename $VMRC` | cut -d. -f1).rc

[ -f "$CFGFILE" ] && . ./$CFGFILE || error 6

#********************************************************************************
[ "$DESC" ] && FULLNAME="$NAME ($DESC)" || FULLNAME="$NAME"
# the pidfile and logfile are located in the VM's folder
PIDFILE=`pwd`/$(echo "$NAME" | tr '[A-Z]' '[a-z]')-qemu.pid
LOGFILE=`pwd`/$(echo "$NAME" | tr '[A-Z]' '[a-z]')-qemu.log

#********************************************************************************
${ACTION}                 # now run the requested action (ok, its a bit bash'ish)
exit 0
